uses crt;

const
  startbank = $300200;
  startofs = 0;
  engsize = 1024;
  nwidth = 225;

type
  japrec = record
    fpos : longint;
    size : integer;
    st : array[0..511] of byte;
  end;
  engrec = record
    size : integer;
    st : array[0..engsize-1] of byte;
  end;
  fontt2 = array[0..102, 0..11] of word;

var
  newfont : ^fontt2;
  widths : array[0..255] of byte;
  txt : array[0..engsize-1] of byte;
  num, sz, x, y, n, i : longint;
  count, offset : word;
  fp, engfp, japfp, rom : file;
  pre, post : string[3];
  engtext : engrec;
  japtext : japrec;
  temp : string;
  ch : char;
  dark : boolean;
  code : byte;

procedure add(c : byte);
begin
  if sz > 0 then
    begin
      if ((c = 1) or (c = 2)) and (txt[sz-1] = 16) then dec(sz);
      if (c = 1) and (txt[sz-1] = 2) then dec(sz);
    end;
  inc(sz);
  txt[sz-1] := c;
end;

procedure printeng(text : engrec);
var
  endwin : boolean;
  ts, l, tn, i, n : integer;
  c, col : byte;
  temp : string;

  procedure dump;
    var q : byte;
    begin
      tn := 0;
      for q := 1 to length(temp) do
        begin
          c := ord(temp[q]);

          if ((not dark) and (c and 128 <> 0))
          or ((dark) and (c and 128 = 0)) then
            begin
              add(8);
              dark := not dark;
            end;

          add(c and 127);
          inc(n, widths[c and 127]);
        end;
      temp := '';
    end;

begin
  tn := 0;
  temp := '';
  l := 0;
  n := 0;
  endwin := false;
  for i := 0 to text.size-1 do
    begin
      c := text.st[i] and 127;
      if (c >= 16) and (tn+widths[c] > nwidth) then { where's the space! }
        begin
          dump;
          add(2-ord(endwin));
          endwin := not endwin;
          n := 0;
        end;
      if text.st[i] <> 2 then
        temp := temp+chr(text.st[i]);
      inc(tn, widths[c]);
      if c = 16 then
        dump
      else if n+tn > nwidth then
        begin
          add(2-ord(endwin));
          endwin := not endwin;
          n := 0;
        end;
      if c = 2 then
        begin
          dump;
          add(2-ord(endwin));
          endwin := not endwin;
          n := 0;
        end;
    end;
  dump;
end;

function totext(c : byte) : string;
begin
  case c of
    16..31 : totext := chr(c+16);
    32..38 : totext := chr(c+26);
    39..44 : totext := chr(c+52);
    45..48 : totext := chr(c+78);
    49..50 : totext := '"';
    51 : totext := '...';
    52 : totext := '[Heart]';
    53 : totext := '[Star]';
    54 : totext := '[Circle]';
    55 : totext := '[Triangle]';
    56 : totext := '[Note]';
    57..66 : totext := chr(c-9);
    67..92 : totext := chr(c-2);
    93..118 : totext := chr(c+4);
  else
    totext := '<'+chr(c+48)+'>';
  end;
end;

begin
  if paramcount <> 1 then
    begin
      writeln('Usage: TEXTINS rom.smc');
      exit;
    end;

  assign(rom, paramstr(1));
  reset(rom, 1);

  filemode := 0;

  new(newfont);
  assign(fp, 'mainfont.dat');
  reset(fp, 1);
  blockread(fp, newfont^, sizeof(newfont^));
  close(fp);
  for i := 0 to 102 do
    for y := 0 to 11 do
      newfont^[i, y] := swap(newfont^[i, y]);

  fillchar(widths, sizeof(widths), 0);
  for i := 0 to 102 do
    begin
      n := 0; x := 16;
      for y := 0 to 11 do
        n := n or newfont^[i, y];
      while (n and 1 <> 1) and (x <> 0) do
        begin
          dec(x);
          n := n shr 1;
        end;
      if x = 0 then
        x := 5;
      widths[i+16] := x+1;
    end;
  widths[6] := widths[89]*7;
  widths[7] := widths[89]*7;
  widths[9] := widths[61]*3;
  dispose(newfont);

  assign(engfp, 'engtext.dat');
  reset(engfp, 1);
  blockread(engfp, count, sizeof(count));
  assign(japfp, 'japtext.dat');
  reset(japfp, 1);
  blockread(japfp, count, sizeof(count));

  code := 3;
  offset := startofs;
  num := 0;
  repeat
    blockread(engfp, engtext, sizeof(engtext));
    blockread(japfp, japtext, sizeof(japtext));

    pre := ''; post := '';
    if japtext.st[0] = 5 then
      pre := #5+chr(japtext.st[1]);
    if japtext.st[japtext.size-3] = 11 then
      post := #11#1#0
    else post := #1#0;

    with engtext do
      begin
        sz := 0;
        for n := 1 to length(pre) do
          begin
            txt[sz] := ord(pre[n]);
            inc(sz);
          end;
        dark := false;
        printeng(engtext);
        if dark then add(8);
        for n := 1 to length(post) do
          begin
            txt[sz] := ord(post[n]);
            inc(sz);
          end;
      end;
    seek(rom, japtext.fpos);
    if sz > japtext.size then
      begin
        blockwrite(rom, txt, japtext.size-3);
        blockwrite(rom, code, 1);
        blockwrite(rom, offset, 2);
        seek(rom, startbank+longint(offset));
        blockwrite(rom, txt[japtext.size-3], sz-japtext.size+3);
        offset := offset+sz-japtext.size+3;
      end
    else blockwrite(rom, txt, sz);
    inc(num);
  until eof(engfp);

  close(rom);
  close(engfp);
end.
