uses crt, dos;

const
  { placeholder strings }
  PlayerStr = 'Player';
  PinoStr   = 'Pino';
  NumberStr = '999';


  engsize = 1024; { maximum chars allowed in a page }
  sxold = 52;
  syold = 19;
  sxnew = 48;
  synew = 106;
  nwidth = 225;
  owidth = 12*18;
  arrow : array[0..6, 0..6] of byte =
  ((1,1,1,0,1,1,1),
   (1,1,0,8,0,1,1),
   (1,0,8,8,8,0,1),
   (0,8,8,8,8,8,0),
   (0,0,0,0,0,0,0),
   (0,8,8,8,8,8,0),
   (0,0,0,0,0,0,0));
  paragraph : array[0..11] of byte =
  (0, 124, 244, 244, 244, 244, 116, 20, 20, 20, 20, 0);
  space : array[0..11] of byte =
  (0, 0, 0, 0, 0, 96, 96, 0, 0, 0, 0, 0);

type
  japrec = record
    fpos : longint;
    size : integer;
    st : array[0..511] of byte;
  end;
  engrec = record
    size : integer;
    st : array[0..engsize-1] of byte;
  end;
  fontt1 = array[0..455, 0..11] of word;
  fontt2 = array[0..102, 0..11] of word;
  screen = array[0..199, 0..319] of byte;
  printproc = procedure(x, l, pos : integer; c : byte);

var
  KeyStatus : Byte Absolute $40:$17;

  regs : registers;
  japfp, engfp, fp : file;
  oldfont : ^fontt1;
  newfont : ^fontt2;
  widths : array[0..255] of byte;
  back, scrn : ^screen;
  maxline, japline, engline, cpos : integer;
  x, y, i, maxchar, maxdigit : integer;
  japtext : japrec;
  engtext : engrec;
  refjap, sptoggle, ptoggle : boolean;
  num, n : longint;
  count : word;
  japcol : byte;

  ch : char;

  cx, cline, sx, sline : integer; { used for cursor seeks }

procedure color(n, r, g, b : byte);
begin
  port[$3C6] := $FF;
  port[$3C8] := n;
  port[$3C9] := r shr 2;
  port[$3C9] := g shr 2;
  port[$3C9] := b shr 2;
end;

procedure readcolors;
var
  txt : text;
  r, g, b, n : byte;
begin
  assign(txt, 'trans.pal');
  reset(txt);
  readln(txt); readln(txt); readln(txt);
  for n := 0 to 255 do
    begin
      readln(txt, r, g, b);
      color(n, r, g, b);
    end;
  close(txt);
end;

function kbtochar(ch : char) : byte;
var c : byte;
begin
  c := ord(ch);
  case ch of
    #24..#31 : kbtochar := c+9;   { 21..28 }
    ' '..'/' : kbtochar := c-32;  { 00..0F }
    '0'..'9' : kbtochar := c-7;   { 29..32 }
    ':'..'@' : kbtochar := c-42;  { 10..16 }
    'A'..'Z' : kbtochar := c-14;  { 33..4C }
    '['..'`' : kbtochar := c-68;  { 17..1C }
    'a'..'z' : kbtochar := c-20;  { 4D..66 }
    '{'..'~' : kbtochar := c-94;  { 1D..20 }
  else
    kbtochar := 0;
  end;
end;

function totext(c : byte) : string;
begin
  if c < 16 then
    begin
      case c of
        2 : totext := #13#10;
        6 : totext := '[Player]';
        7 : totext := '[Pino]';
        9 : totext := '[Number]';
      end;
    end
  else begin
    dec(c, 16);
    case c of
      $00..$06 : totext := chr(c+32);  { ' '..'&' }
      $07 : totext := #146;            { right single quote }
      $08..$0F : totext := chr(c+32);  { '('..'/' }
      $10..$16 : totext := chr(c+42);  { ':'..'@' }
      $17..$1B : totext := chr(c+68);  { '['..'_' }
      $1C : totext := #145;            { left single quote }
      $1D..$20 : totext := chr(c+94);  { '{'..'~' }
      $21 : totext := #147;            { left quote }
      $22 : totext := #148;            { right quote }
      $23 : totext := #133;            { ellipsis }
      $24 : totext := '[Heart]';       { heart }
      $25 : totext := '[Star]';        { star }
      $26 : totext := '[Circle]';      { circle }
      $27 : totext := '[Triangle]';    { triangle }
      $28 : totext := '[Note]';        { note }
      $29..$32 : totext := chr(c+7);   { '0'..'9' }
      $33..$4C : totext := chr(c+14);  { 'A'..'Z' }
      $4D..$66 : totext := chr(c+20);  { 'a'..'z' }
    end;
  end;
end;

function tohex(x : longint) : string;
  const hex : array[0..15] of char = ('0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F');
  var st : string;
  begin
    st := '';
    repeat
      st := hex[x and 15]+st;
      x := x shr 4;
    until x = 0;
    tohex := st;
  end;

procedure printst(x1, y1 : integer; st : string; col : byte);
var
  n, x, y : integer;
  c : byte;
begin
  for n := 1 to length(st) do
    begin
      c := kbtochar(st[n]);
      for y := 0 to 11 do
      for x := 0 to widths[c+16]-2 do
        if newfont^[c, y] and (32768 shr x) <> 0 then
          scrn^[y+y1,x+x1] := col;
      x1 := x1+widths[c+16];
    end;
end;

procedure printjap(text : japrec);
var
  n, i, sy, l : integer;
  x, y, col : byte;
  c : word;

  procedure setlines(s : integer);
  var y : integer;
  begin
    if n+s > sxold+owidth then
      begin
        n := sxold;
        inc(l);
      end;

    if l > maxline then
      maxline := l;
    if l < japline then
      for y := 0 to 6 do
        move(arrow[y], scrn^[syold+y-5, 285], 7);
    if l > japline+1 then
      for y := 0 to 6 do
        move(arrow[y], scrn^[syold-y+32, 285], 7);
  end;

begin
  n := sxold;
  l := 0;
  i := 0;
  japcol := 0;
  col := 0;
  repeat
    c := text.st[i];
    if c = 255 then
      begin
        inc(i);
        c := 255+text.st[i];
      end;
    if c >= 16 then
      begin
        dec(c, 16);
        setlines(12);
        if (l = japline) or (l = japline+1) then
          begin
            sy := (l-japline)*16+syold;
            for y := 0 to 11 do
            for x := 0 to 11 do
              if oldfont^[c, y] and (32768 shr x) <> 0 then
                scrn^[y+sy,x+n] := col+8;
            n := n+12;
          end;
      end
    else begin
      case c of
        1 : begin n := sxold; inc(l); if odd(l) then inc(l); end;
        2 : begin n := sxold; inc(l); end;
        5 : begin inc(i); col := text.st[i]; japcol := col; end;
        6 : if (l = japline) or (l = japline+1) then
              begin
                setlines(48);
                sy := (l-japline)*16+syold;
                printst(n+1, sy+1, PlayerStr, col+8+4);
                printst(n, sy, PlayerStr, col+8);
                inc(n, 48);
              end;
        7 : if (l = japline) or (l = japline+1) then
              begin
                setlines(48);
                sy := (l-japline)*16+syold;
                printst(n+1, sy+1, PinoStr, col+8+4);
                printst(n, sy, PinoStr, col+8);
                inc(n, 48);
              end;
        8 : col := (col+4) and 7;
        9 : if (l = japline) or (l = japline+1) then
              begin
                setlines(24);
                sy := (l-japline)*16+syold;
                printst(n+1, sy+1, NumberStr, col+8+4);
                printst(n, sy, NumberStr, col+8);
                inc(n, 24);
              end;
      end;
    end;
    inc(i);
  until i = text.size;
end;

function pos(substr : string; text : engrec) : integer;
  var p, i : integer;
  begin
    p := 0;
    repeat
      while (text.st[p] and 127 <> ord(substr[1])) and (p < text.size) do
        inc(p);
      i := 1;
      while (text.st[p+i] and 127 = ord(substr[1+i])) and (i < length(substr)) and (p+i < text.size) do
        inc(i);
      inc(p);
    until (p > text.size) or (i >= length(substr));

    if p <= text.size then
      pos := p
    else pos := 0;
  end;

procedure delete(var text : engrec; index : integer);
  var i : longint;
  begin
    if index <= text.size then
      begin
        for i := index-1 to longint(text.size)-2 do
          text.st[i] := text.st[i+1];
        dec(text.size);
        text.st[text.size] := 0;
      end;
  end;

procedure insert(b : byte; var text : engrec; index : integer);
  var i : word;
  begin
    if index <= text.size+1 then
      begin
        for i := text.size downto index do
          text.st[i] := text.st[i-1];
        text.st[index-1] := b;
        inc(text.size);
      end;
  end;

procedure clean(var text : engrec);
  var p, n : integer;
  begin
    p := 0;
    for n := 0 to text.size-1 do 
      begin
        if (text.st[n] and 127) in [18, 49, 50] then { fix quotes }
          begin
            text.st[n] := 49+p+text.st[n] and 128;
            p := 1-p;
          end;
        if (n > 0) then { fix any speed oddities in whitespace }
          if (text.st[n] and 127) in [2, 16] then
            text.st[n] := (text.st[n] and 127) or (text.st[n-1] and 128);
      end;
    p := pos(#16#16, text); { remove double spaces }
    while p <> 0 do
      begin
        if cpos = p+2 then
          cpos := cpos-1;
        delete(text, p);
        p := pos(#16#16, text);
      end;
    p := pos(#30#30, text); { remove double periods }
    while p <> 0 do
      begin
        if cpos = p+2 then
          cpos := cpos-1;
        delete(text, p);
        p := pos(#30#30, text);
      end;
    n := 1;
    if cpos < 1 then
      cpos := 1;
    if cpos > text.size+1 then
      cpos := text.size+1;
  end;

procedure printchar(x1, l, pos : integer; c : byte); far;
var x, y, y1, col : byte;
begin
  y1 := synew+(l-engline)*16;
  col := japcol+(c shr 7)*4;
  if c and 127 >= 16 then
    begin
      if (l = engline) or (l = engline+1) then
        begin
          c := (c and 127)-16;

          if c <> 0 then
            begin
              for y := 0 to 11 do
              for x := 0 to widths[c+16]-2 do
                if newfont^[c, y] and (32768 shr x) <> 0 then
                  scrn^[y+y1,x+x1] := col+8;
            end
          else if ptoggle then
            begin
              for y := 0 to 11 do
              for x := 0 to 7 do
                if space[y] and (128 shr x) <> 0 then
                  scrn^[y+y1,x+x1] := col+8;
            end;
        end;
    end
  else if (l = engline) or (l = engline+1) then
    begin
      case c and 127 of
        2 : if ptoggle then
              begin
                for y := 0 to 11 do
                for x := 0 to 7 do
                  if paragraph[y] and (128 shr x) <> 0 then
                    scrn^[y+y1,x+x1] := col+8;
              end;
        6 : begin
              printst(x1+1, y1+1, PlayerStr, col+8+4);
              printst(x1, y1, PlayerStr, col+8);
            end;
        7 : begin
              printst(x1+1, y1+1, PinoStr, col+8+4);
              printst(x1, y1, PinoStr, col+8);
            end;
        9 : begin
              printst(x1+1, y1+1, NumberStr, col+8+4);
              printst(x1, y1, NumberStr, col+8);
            end;
      end;
    end;
  if l < engline then
    for y := 0 to 6 do
      move(arrow[y], scrn^[synew+y-5, 285], 7);
  if l > engline+1 then
    for y := 0 to 6 do
      move(arrow[y], scrn^[synew-y+32, 285], 7);
end;

procedure drawcursor(x, l, pos : integer; c : byte); far;
var n, w : integer;
begin
  c := c and 127;
  if pos = cpos then
    begin
      w := widths[c];
      if w = 0 then
        w := 7;
      if (x < sxnew+230) and ((l = engline) or (l = engline+1)) then
        for n := 0 to w-2 do
          scrn^[synew+(l-engline)*16+12, x+n] := japcol+ord(sptoggle)*4+8;
      cx := x;
      cline := l;
    end;
end;

procedure findcursor(x, l, pos : integer; c : byte); far;
begin
  if pos = 1 then
    cx := x;
  if (l = sline) and (abs(sx-x) <= abs(sx-cx)) then
    begin
      cx := x;
      cpos := pos;
    end;
end;

procedure printeng(text : engrec; proc : printproc);
var
  ts, l, tn, n, i, sy, cx, cy : integer;
  c, col : byte;
  temp : string;

  procedure dump;
    var q : byte;
    begin
      tn := 0;
      for q := 1 to length(temp) do
        begin
          c := ord(temp[q]);
          proc(n, l, ts, c);
          c := c and 127;
          inc(n, widths[c]);
          inc(ts, 1);
        end;
      temp := '';
    end;

begin
  n := sxnew;
  tn := 0;
  temp := '';
  ts := 1;
  l := 0;
  for i := 1 to text.size do
    begin
      c := text.st[i-1] and 127;
      if (c >= 16) and (tn+widths[c] > nwidth) then { where's the space! }
        begin
          dump;
          n := sxnew;
          inc(l);
        end;
      temp := temp+chr(text.st[i-1]);
      inc(tn, widths[c]);
      if c = 16 then
        dump
      else if n+tn > sxnew+nwidth then
        begin
          n := sxnew;
          inc(l);
        end;
      if c = 2 then
        begin
          dump;
          n := sxnew;
          inc(l);
        end;
    end;
  temp := temp+#15;
  dump;
end;

procedure clearwindows(b : byte);
var y : integer;
begin
  for y := 0 to 37 do
    begin
      if (b and 1 <> 0) then
        move(back^[y+syold-5, sxold], scrn^[y+syold-5, sxold], 300-sxold);
      if (b and 2 <> 0) then
        move(back^[y+synew-5, sxnew], scrn^[y+synew-5, sxnew], 300-sxnew);
    end;
end;

procedure saveeng;
  var p : integer;
begin
  { Do a little cleaning before saving }
  p := pos(#16#2, engtext); { remove space enter }
  while p <> 0 do
    begin
      if cpos = p+2 then
        cpos := cpos-1;
      delete(engtext, p);
      p := pos(#16#2, engtext);
    end;
  while (engtext.size <> 0) and (engtext.st[engtext.size-1] = 2) do
    begin { remove excess newline }
      if cpos = engtext.size+2 then
        cpos := cpos-1;
      delete(engtext, engtext.size);
    end;

  seek(engfp, num*sizeof(engtext)+sizeof(count));
  blockwrite(engfp, engtext, sizeof(engtext));
end;

procedure changepage(n : integer);
var
  st : string;
begin
  move(back^, scrn^, sizeof(back^));

  num := n;
  seek(engfp, num*sizeof(engtext)+sizeof(count));
  blockread(engfp, engtext, sizeof(engtext));
  seek(japfp, num*sizeof(japtext)+sizeof(count));
  blockread(japfp, japtext, sizeof(japtext));
  sptoggle := false;
  cpos := 1;
  engline := 0;
  japline := 0;
  maxline := 0;
  refjap := true;

  str(num, st);
  printst(252, 62, st, 8);
end;

procedure jumppage;
var
  ch : char;
  n : integer;
  st : string;
begin
  n := num;
  repeat
    changepage(num);
    str(num, st);
    st := st+'_';
    printst(252, 62, st, 8);
    clearwindows(3);
    printjap(japtext);
    printeng(engtext, printchar);
    move(scrn^, mem[$A000:0], 64000);
    ch := readkey;
    case ch of
      #0 : readkey;
      #8 : num := num div 10;
      '0'..'9' : num := num*10+ord(ch)-48;
    end;
    if num > count-1 then
      num := count-1;
  until (ch = #13) or (ch = #27);
  if ch = #27 then
    num := n;
  changepage(num);
end;

procedure jumpempty;
var
  n : longint;
  engtext : engrec;
begin
  n := num;
  repeat
    inc(n);
    if n < count then
      begin
        seek(engfp, n*sizeof(engtext)+sizeof(count));
        blockread(engfp, engtext, sizeof(engtext));
      end;
  until (n >= count) or (engtext.size = 0);
  if n >= count then
    n := count-1;
  changepage(n);
end;

procedure dumpeng;
var
  engtext : engrec;
  n, i : integer;
  fp : text;
  count : word;
begin
  assign(fp, 'textdump.txt');
  rewrite(fp);
  i := 0;
  seek(engfp, 0);
  blockread(engfp, count, sizeof(count));
  writeln(fp, '--Count ', count, '--');
  writeln(fp);
  repeat
    blockread(engfp, engtext, sizeof(engtext));
    if engtext.size > 0 then
      begin
        writeln(fp, '--Page ', i, '--');
        if engtext.st[0] and 128 <> 0 then
          write(fp, '[8]');
        for n := 0 to engtext.size-1 do
          begin
            if (n <> 0) and ((engtext.st[n-1] and 128) <> (engtext.st[n] and 128)) then
              write(fp, '[8]');
            write(fp, totext(engtext.st[n] and 127));
          end;
        if engtext.st[engtext.size-1] and 128 <> 0 then
          write(fp, '[8]');
        writeln(fp);
        writeln(fp);
      end;
    inc(i);
  until eof(engfp);
  writeln(fp, '--End--');
  close(fp);
end;

begin
  clrscr;

  new(oldfont);
  new(newfont);
  new(back); new(scrn);
  assign(fp, 'japfont.dat');
  reset(fp, 1);
  blockread(fp, oldfont^, sizeof(oldfont^));
  close(fp);

  assign(fp, 'mainfont.dat');
  reset(fp, 1);
  blockread(fp, newfont^, sizeof(newfont^));
  close(fp);
  for i := 0 to 102 do
    for y := 0 to 11 do
      newfont^[i, y] := swap(newfont^[i, y]);

  maxchar := 0; maxdigit := 0;
  fillchar(widths, sizeof(widths), 0);
  for i := 0 to 102 do
    begin
      n := 0; x := 16;
      for y := 0 to 11 do
        n := n or newfont^[i, y];
      while (n and 1 <> 1) and (x <> 0) do
        begin
          dec(x);
          n := n shr 1;
        end;
      if x = 0 then { space char defaults to 5 }
        x := 5;
      widths[i+16] := x+1;

      if widths[i+16] > widths[maxchar+16] then { maxchar is usually W }
        maxchar := i;
      if (totext(i+16) >= '0') and (totext(i+16) <= '9') and
        (widths[i+16] > widths[maxdigit+16]) then { maxdigit is usually 4 }
          maxdigit := i;
    end;

  { Reserve some space for large names and numbers }
  widths[6] := widths[maxchar+16]*7;   { Player  max 7 char }
  widths[7] := widths[maxchar+16]*7;   { Pino    max 7 char }
  widths[9] := widths[maxdigit+16]*3;  { Number  max 3 char }

  assign(fp, 'back.raw');
  reset(fp, 1);
  blockread(fp, back^, sizeof(back^));
  close(fp);

  assign(japfp, 'japtext.dat');
  reset(japfp, 1);
  blockread(japfp, count, sizeof(count));

  {$I-}
  assign(engfp, 'engtext.dat');
  reset(engfp, 1);
  if ioresult <> 0 then
    begin
      writeln('Engtext.dat not found. Creating a new one.');
      write('Press any key');
      readkey;
      rewrite(engfp, 1);
      blockwrite(engfp, count, sizeof(count));
      repeat
        blockread(japfp, japtext, sizeof(japtext));
        fillchar(engtext, sizeof(engtext), 0);
        blockwrite(engfp, engtext, sizeof(engtext));
      until eof(japfp);
      seek(japfp, sizeof(count));
    end;
  {$I+}

  regs.ax := $13;
  intr($10, regs);
  fillchar(scrn^, sizeof(scrn^), 0);

  readcolors;
  ptoggle := false;
  changepage(0);

  repeat
    clearwindows(2+ord(refjap));

    if refjap then { japanese window needs a refresh }
      printjap(japtext);
    refjap := false;

    clean(engtext);
    printeng(engtext, printchar);
    printeng(engtext, drawcursor);
    if (cline < engline) or (cline > engline+1) then
      begin { cursor not showing so adjust the engline }
        clearwindows(2);
        while (cline < engline) do dec(engline,2);
        while (cline > engline+1) do inc(engline,2);
        printeng(engtext, printchar);
        printeng(engtext, drawcursor);
      end;
    move(scrn^, mem[$A000:0], 64000);
    ch := readkey;
    case ch of
      #0 : { Function keys }
        begin
          ch := readkey;
          case ch of
            ';'..'@' : { F1..F6 }
              if engtext.size < engsize then
                begin
                  insert(ord(ch)-24+16+ord(sptoggle) shl 7, engtext, cpos);
                  cpos := cpos+1;
                end;
            'A'..'C' : { F7..F9 }
              if engtext.size < engsize then
                begin
                  if ch = 'C' then ch := 'D';
                  insert(ord(ch)-59+ord(sptoggle) shl 7, engtext, cpos);
                  cpos := cpos+1;
                end;
            'H' : { Up }
              begin
                sx := cx;
                sline := cline-1;
                printeng(engtext, findcursor);
              end;
            'P' : { Down }
              begin
                sx := cx;
                sline := cline+1;
                printeng(engtext, findcursor);
              end;
            'G' : { Home }
              begin
                sx := 0;
                sline := cline;
                printeng(engtext, findcursor);
              end;
            'O' : { End }
              begin
                sx := 1000;
                sline := cline;
                printeng(engtext, findcursor);
              end;
            'K' : { Left }
              if cpos > 1 then
                dec(cpos);
            'M' : { Right }
              if cpos <= engtext.size then
                inc(cpos);
            'S' : { Delete }
              if cpos <= engtext.size then
                delete(engtext, cpos);
            #132 : { Ctrl+PgUp }
              begin
                if japline > 0 then dec(japline, 2);
                refjap := true;
              end;
            #118 : { Ctrl+PgDn }
              begin
                if japline < maxline-1 then inc(japline, 2);
                refjap := true;
              end;
            'I' : { Page Up }
              begin
                sx := cx;
                sline := cline-1;
                printeng(engtext, findcursor);
                printeng(engtext, drawcursor);
                sline := cline-1;
                printeng(engtext, findcursor);
                printeng(engtext, drawcursor);
              end;
            'Q' : { Page Down }
              begin
                sx := cx;
                sline := cline+1;
                printeng(engtext, findcursor);
                printeng(engtext, drawcursor);
                sline := cline+1;
                printeng(engtext, findcursor);
                printeng(engtext, drawcursor);
              end;
            #15 : { Shift+Tab }
              if num > 0 then
                begin
                  saveeng;
                  changepage(num-1);
                end;
          end;
        end;
      #4 : { Ctrl+D }
        begin
          saveeng;
          dumpeng;
        end;
      #5 : { Ctrl+E }
        begin
          saveeng;
          jumpempty;
        end;
      #8 : { Backspace }
        if cpos > 1 then
          begin
            cpos := cpos-1;
            delete(engtext, cpos);
          end;
      #9 : { Tab }
        if num < count-1 then
          begin
            saveeng;
            changepage(num+1);
          end;
      #10 : { Ctrl+J }
        jumppage;
      #13 : { Enter }
        if engtext.size < engsize then
          begin
            insert(2+ord(sptoggle) shl 7, engtext, cpos);
            cpos := cpos+1;
          end;
      #16 : { Ctrl+P }
        ptoggle := not ptoggle;
      #19 : { Ctrl+S}
        sptoggle := not sptoggle;
      ' '..'~' : { Printable character }
        if engtext.size < engsize then
          begin
            insert(kbtochar(ch)+16+ord(sptoggle) shl 7, engtext, cpos);
            cpos := cpos+1;
          end;
    end;
  until ch = #27;
  saveeng;
  close(japfp);
  close(engfp);

  dispose(newfont);
  dispose(oldfont);
  dispose(back); dispose(scrn);
  textmode(3);
end.